#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/mman.h>
  
#define FATAL do { fprintf(stderr, "Error at line %d, file %s (%d) [%s]\n", \
  __LINE__, __FILE__, errno, strerror(errno)); exit(1); } while(0)


void ppx2rgb(char *obp, char Y, char U, char V) {
   int C = Y - 16;
   int D = U - 128;
   int E = V - 128;

   int r = ( 298 * C           + 409 * E + 128) >> 8;
   int g = ( 298 * C - 100 * D - 208 * E + 128) >> 8;
   int b = ( 298 * C + 516 * D           + 128) >> 8;

   r = (r < 0)? 0: r; r = (r > 255)? 255: r;
   g = (g < 0)? 0: g; g = (g > 255)? 255: g;
   b = (b < 0)? 0: b; b = (b > 255)? 255: b;

   obp[0]=(unsigned char)r;
   obp[1]=(unsigned char)g;
   obp[2]=(unsigned char)b;
}

/*
void ppx2rgb(char *obp, char y, char cb, char cr) {
   float Y = (float) y;
   float Cb = (float) cb;
   float Cr = (float) cr;

   int r = (int) (Y + 1.40200 * (Cr - 0x80));
   int g = (int) (Y - 0.34414 * (Cb - 0x80) - 0.71414 * (Cr - 0x80));
   int b = (int) (Y + 1.77200 * (Cb - 0x80));

   r = (r < 0)? 0: r; r = (r > 255)? 255: r;
   g = (g < 0)? 0: g; g = (g > 255)? 255: g;
   b = (b < 0)? 0: b; b = (b > 255)? 255: b;

   obp[0]=(unsigned char)r;
   obp[1]=(unsigned char)g;
   obp[2]=(unsigned char)b;
}
*/

void convert_s422(unsigned char *ybp_i, unsigned char *cbp_i, unsigned char *obp) {
   int lc;
   for (lc=0; lc < 1080; lc++) {
      unsigned char *ybp=ybp_i; unsigned char *cbp=cbp_i; int cc;
      for (cc=0; cc < 320; cc++) {
         unsigned char y1,y2,y3,y4,y5,y6;
         unsigned char cb12,cr12,cb34,cr34,cb56,cr56;
         unsigned char b1,b2,b3,b4,b5,b6,b7,b8;

         b1=ybp[3]; b2=ybp[2]; b3=ybp[1]; b4=ybp[0];
         b5=ybp[7]; b6=ybp[6]; b7=ybp[5]; b8=ybp[4];
         y3=((b1 & 0x3f) << 2) | ((b2 & 0xc0) >> 6);
         y2=((b2 & 0x0f) << 4) | ((b3 & 0xf0) >> 4);
         y1=((b3 & 0x03) << 6) | ((b4 & 0xfc) >> 2);
         y6=((b5 & 0x0f) << 4) | ((b6 & 0xf0) >> 4);
         y5=((b6 & 0x03) << 6) | ((b7 & 0xfc) >> 2);
         y4=b8;

         b1=cbp[3]; b2=cbp[2]; b3=cbp[1]; b4=cbp[0];
         b5=cbp[7]; b6=cbp[6]; b7=cbp[5]; b8=cbp[4];
         cb34=((b1 & 0x3f) << 2) | ((b2 & 0xc0) >> 6);
         cr12=((b2 & 0x0f) << 4) | ((b3 & 0xf0) >> 4);
         cb12=((b3 & 0x03) << 6) | ((b4 & 0xfc) >> 2);
         cr56=((b5 & 0x0f) << 4) | ((b6 & 0xf0) >> 4);
         cb56=((b6 & 0x03) << 6) | ((b7 & 0xfc) >> 2);
         cr34=b8;

         ppx2rgb(obp, y1, cb12, cr12); obp+=3;
         ppx2rgb(obp, y2, cb12, cr12); obp+=3;
         ppx2rgb(obp, y3, cb34, cr34); obp+=3;
         ppx2rgb(obp, y4, cb34, cr34); obp+=3;
         ppx2rgb(obp, y5, cb56, cr56); obp+=3;
         ppx2rgb(obp, y6, cb56, cr56); obp+=3;

         ybp+=8; cbp+=8;
      }
      ybp_i+=2624; cbp_i+=2624;
   }
}

int dumptofile(int wfd, char *y_ptr, char *c_ptr, unsigned long msize) {
   int outsize=1920*1080*3;

   void *y_buf=malloc(msize); if (y_buf == NULL) {
      fprintf(stderr, "\nMalloc error (Y buf)!!!\n"); fflush(stdout);
      return(1); }
   void *c_buf=malloc(msize); if (c_buf == NULL) {
      fprintf(stderr, "\nMalloc error (C buf)!!!\n"); fflush(stdout);
      return(1); }
   void *outbuf=malloc(outsize); if (outbuf == NULL) {
      fprintf(stderr, "\nMalloc error (outbuf)!!!\n"); fflush(stdout);
      return(1); }

   memcpy(y_buf, y_ptr, msize);
   memcpy(c_buf, c_ptr, msize);
   // memset(outbuf, 0, outsize);

   printf("Converting...\n"); fflush(stdout);
   convert_s422(y_buf, c_buf, outbuf);

   printf("Writing %d bytes...", outsize); fflush(stdout);
   if (write(wfd, outbuf, outsize) != outsize) {
      fprintf(stderr, "\nWrite error !!!\n"); fflush(stdout); }

   printf("%s", " done.\n"); fflush(stdout);
   free(y_buf); free(c_buf); free(outbuf);
   return(0);
}

int main(int argc, char **argv) {
   int fd, fd2, wrfd;

   if (argc != 2) {
      fprintf(stderr, "\nUsage:\t%s file.rgb\n\n", argv[0]);
      exit(1); }

   // LE??B650T2W; F/W: TCHL7DEUC-2005.0; input: HDMI1-4, 1920x1080@p24/p50/p60
   // Y (Y' - ?) planes: 0x6c400000, 0x6c6d2000, 0x6cc76000, 0x6cf48000
   off_t target=0x6c6d2000;
   int bcount=2833920;

   int map_size=bcount/4096; map_size*=4096;
   if (map_size < bcount) map_size+=4096;
   if (target % 4096 || bcount < 1 || bcount > 0x4000000) {
      fprintf(stderr, "Address / size not valid (Y)\n");
      exit(1);
   }
   if ((fd=open("/dev/mem", O_RDWR | O_SYNC)) == -1) FATAL;
   char *plane_y=mmap(0, map_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, target);
   if (plane_y == (void *) -1) FATAL;
   printf("Plane Y (%p) mapped at address %p; size=%d\n", target, plane_y, map_size);

   // CbCr (4:2:2) planes: 0x6d7be000, 0x6da90000, 0x6dd62000, 0x6e034000, 0x6e306000 (???)
   target=0x6da90000;
   if (target % 4096 || bcount < 1 || bcount > 0x4000000) {
      fprintf(stderr, "Address / size not valid (CrCb)\n");
      exit(1);
   }
   if ((fd2=open("/dev/mem", O_RDWR | O_SYNC)) == -1) FATAL;
   char *plane_c=mmap(0, map_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd2, target);
   if (plane_c == (void *) -1) FATAL;
   printf("Plane CbCr (%p) mapped at address %p; size=%d\n", target, plane_c, map_size);

   if ((wrfd=open(argv[1], O_WRONLY | O_CREAT | O_TRUNC, 0666)) == -1) FATAL;
   printf("File '%s' opened for writing.\n",argv[1]);
   fflush(stdout);

   dumptofile(wrfd, plane_y, plane_c, bcount);

   if (munmap(plane_y, map_size) == -1) FATAL;
   if (munmap(plane_c, map_size) == -1) FATAL;
   close(fd); close(fd2); close(wrfd);
   return 0;
}
